import re
import sys
from ..base.model import Model
from ..core.object import Object
from ..core.class_ import Class
from ..core.function import Function
from ..utils.error import Error
from ..utils.common import parse_object, smart_split


class Parser(object):

    def __init__(self, model):
        self.model = model

    def parse_files(self, files):
        for path in files:
            text = open(path, encoding='utf-8').read()
            self.parse_text(text)

    def parse_text(self, text):
        text = self.remove_comments(text)
        while text:
            if Parser._is_class(text):
                text = self._create_class(text, False)
            elif Parser._is_interface(text):
                text = self._create_class(text, True)
            elif Parser._is_enum(text):
                text = self._create_enum_class(text)
            elif Parser._is_include(text):
                text = self._create_include(text)
            elif Parser._is_function(text):
                text = self._create_function(text)
            elif Parser._is_constructor(text):
                text = self._create_constructor(text)
            else:
                text = self._create_member(text)
            text = text.strip()

    @staticmethod
    def remove_comments(text):
        left = text.find('/*')
        while left > -1:
            right = text.find('*/', left)
            if right > -1:
                text = text[:left] + text[right + 2:]
            left = text.find('/*')
        lines = text.split('\n')
        for i, line in enumerate(lines):
            if '//' in line:
                lines[i] = line[0:line.find('//')]
        text = '\n'.join(lines).strip()
        return text

    def check_skip(self, line):
        corresponds = True
        first_match = True
        for lang in ['cpp', 'py', 'php', 'js']:
            modifier = ':' + lang
            if modifier in line:
                if first_match:
                    first_match = False
                    corresponds = False
                corresponds = corresponds or self.model.is_lang(lang)
                line = line.replace(modifier, '')
        return not corresponds, line

    def _create_class(self, text, is_abstract):
        body, header, text = Parser.parse_body(text)
        skip, header = self.check_skip(header)
        cls = Class()
        Parser.parse_class(cls, header)
        if is_abstract:
            cls.is_abstract = True

        skip = skip or not self.model.is_side(cls.side)
        if skip:
            if cls.is_storage:
                self.model.classes_for_data.append(cls)
            return text

        parser = Parser(self.model.empty_copy())
        parser.model.side = self.model.side
        parser.parse_class_body(cls, body)
        if self.model.has_class(cls.name):
            Error.exit(Error.DUBLICATE_CLASS, cls.name)
        for inner_cls in cls.inner_classes:
            self.model.add_class(inner_cls)

        cls.generate_constructor()
        self.model.add_class(cls)
        return text

    def _create_enum_class(self, text):
        body, header, text = Parser.parse_body(text)

        skip, header = self.check_skip(header)
        if skip:
            return text

        cls = Class()
        cls.type = 'enum'
        Parser.parse_class(cls, header)
        cls.superclasses.append('BaseEnum')
        if not self.model.is_side(cls.side):
            return text
        parser = Parser(self.model.empty_copy())
        parser.model.side = self.model.side
        parser.parse_class_body(cls, body)
        self.model.add_class(cls)
        return text

    def _create_member(self, text):
        lines = text.split('\n')
        line = lines[0]
        if len(lines) > 1:
            text = text[text.find('\n') + 1:]
        else:
            text = ''

        skip, line = self.check_skip(line)
        if skip:
            return text

        obj = Object()
        self.parse_object(obj, line)
        if self.model.is_side(obj.side):
            self.model.objects.append(obj)
        return text

    def _create_include(self, text):
        k = text.find('\n')
        line = text[0:k]
        text = text[k+1:]
        class_name = line.replace('@include ', '').strip()
        self.model.includes.add(class_name)
        return text

    def _create_function(self, text):
        body, header, text = Parser.parse_body(text)
        skip, header = self.check_skip(header)
        if skip:
            return text
        method = Function()
        self.parse_function_header(method, header)
        if self.model.is_side(method.side):
            method.parse_method_body(body)
            self.model.functions.append(method)
        return text

    def _create_constructor(self, text):
        body, header, text = Parser.parse_body(text)
        skip, header = self.check_skip(header)
        if skip:
            return text
        constructor = Function()
        self.parse_function_header(constructor, 'function void {}'.format(header))
        if self.model.is_side(constructor.side):
            constructor.parse_method_body(body)
            self.model.functions.append(constructor)
        return text

    @staticmethod
    def create_object(description):
        parser = Parser(Model())
        parser.parse_text(description)
        return parser.model.objects[0]

    @staticmethod
    def create_function(description):
        parser = Parser(Model())
        parser.parse_text(description)
        return parser.model.functions[0]

    def parse_object(self, obj, line):
        if ';' in line:
            Error.warning(Error.WARNING_SYNTAX, line)
            line = line.replace(';', '')

        # parse initialize value
        expression = ''
        k = line.find('=')
        if k > -1:
            expression = line[k + 1:].strip()
            line = line[0:k].strip()
        if expression:
            obj.initial_value = expression

        # parse type, name, template arguments
        parse_object(obj, line)

        has_callable, obj.callable_args, obj.type = self.parse_function_args(obj.type)
        obj.callable_args = [a[1] for a in obj.callable_args]
        if not has_callable:
            obj.callable_args = None

        # parse specific modifiers
        obj.type = obj.find_modifiers(obj.type)
        obj.is_pointer = obj.check_pointer()
        obj.is_ref = obj.check_ref()
        if not self.model.is_side(obj.side):
            return

        # recursive parsing template arguments
        args = []
        for arg_desc in obj.template_args:  # type: str
            arg = Object()
            self.parse_object(arg, arg_desc)
            args.append(arg)
        obj.template_args = args

    FUNC_0 = re.compile(r'function<([\w, ]+?)>')
    FUNC_1 = re.compile(r'function<[\w, ]+?>')
    FUNC_0s = re.compile(r'fn<([\w, ]+?)>')
    FUNC_1s = re.compile(r'fn<[\w, ]+?>')
    FUNC_2 = re.compile(r'{.*}')

    def parse_function_header(self, method, line):
        line = line.strip()

        use_s = False
        templates = Parser.FUNC_0.findall(line)
        if not templates:
            templates = Parser.FUNC_0s.findall(line)
            use_s = True
        if templates:
            line = Parser.FUNC_1.sub('', line) if not use_s else Parser.FUNC_1s.sub('', line)
            method.template_types = smart_split(templates[0], ',')
            method.template_types = [x.strip() for x in method.template_types]
        else:
            len_function = 8 if line.startswith('function') else 2 # len('function') or len('fn')
            line = line[len_function:].strip()

        has_callable, method.args, line = self.parse_function_args(line)
        assert has_callable

        line = line.replace(';', '')
        line = Parser.FUNC_2.sub('', line)
        k = line.rfind(' ')

        name_s = line[k:].strip()
        name_s = method.find_modifiers(name_s)
        method.name = name_s

        return_s = line[:k].strip()
        method.return_type = Object()
        self.parse_object(method.return_type, return_s)
        return method

    def parse_function_args(self, line):
        result = []
        left = line.find('(')
        last = line.rfind(')')
        if left == -1 or last == -1 or last < left:
            return False, result, line

        args_s = line[left + 1:last]
        args = smart_split(args_s, ',')

        for arg in args:
            arg = arg.strip()
            if arg.startswith('const '):
                Error.exit(Error.ERROR_CONST_MODIFIER, arg)

            obj = Object()
            self.parse_object(obj, arg)
            result.append([obj.name, obj])

        line = line[:left] + line[last + 1:]
        return True, result, line

    @staticmethod
    def _is_class(line):
        return line.find('class') == 0

    @staticmethod
    def _is_interface(line):
        return line.find('interface') == 0

    @staticmethod
    def _is_function(line):
        return line.find('function') == 0 or line.find('fn') == 0

    @staticmethod
    def _is_constructor(line):
        return line.find('constructor') == 0

    @staticmethod
    def _is_include(line):
        return line.startswith('@include')

    @staticmethod
    def _is_enum(line):
        return line.find('enum') == 0

    @staticmethod
    def parse_body(text):
        text = text.strip()
        body = ''
        k = text.find('\n')
        if k > -1:
            header = text[0:k]
        else:
            header = text
        if header.find(':external') == -1 and header.find(':abstract') == -1:
            text = text[text.find('{'):]
            counter = 0
            left = 0
            last = 0
            for char in text:
                last += 1
                if not counter and char == '{':
                    counter += 1
                    left = last
                    continue

                if char == '{':
                    counter += 1
                if char == '}':
                    counter -= 1
                if not counter:
                    body = text[left:last-1]
                    text = text[last:]
                    break
        else:
            text = text[len(header):].strip()
        return body, header, text

    def load_default_serialize_protocol(self, text):
        lines = text.split('\n')
        supported_types = []
        supported_types.extend(self.model.simple_types)
        supported_types.append('serialized')
        supported_types.append('pointer')
        supported_types.append('list<serialized>')
        supported_types.append('list<pointer>')
        supported_types.append('link')
        supported_types.append('list<link>')
        supported_types.append('serialized')
        supported_types.append('map')
        supported_types.append('enum')
        supported_types.append('list<enum>')
        for type_ in self.model.simple_types:
            list_type = "list<{0}>".format(type_)
            supported_types.append(list_type)

        serialize_protocol = list()
        serialize_protocol.append({})
        serialize_protocol.append({})
        for index in range(2):
            for type_ in supported_types:
                simple = type_ in self.model.simple_types
                protocol_0 = self._load_protocol(lines, index, type_, True if simple else None,
                                                 optional=type_ == 'list<enum>')
                protocol_1 = protocol_0 if not simple else self._load_protocol(lines, index, type_, False)
                serialize_protocol[index][type_] = []
                serialize_protocol[index][type_].extend([protocol_0, protocol_1])
        self.model.serialize_protocol = serialize_protocol

    @staticmethod
    def _load_protocol(lines, serialize_type, type_object, initial_value=None, optional=False):
        serialize_types = ['serialize', 'deserialize']
        serialize_init = ['with default value', 'without default value']
        in_type = False
        in_serialize = False
        in_initial_value = initial_value is None
        pattern = []
        if type_object in ['int64_t', 'uint', 'uint64_t', 'unsigned']:
            type_object = 'int'
        if type_object in ['double']:
            type_object = 'float'
        if type_object in ['list<int64_t>', 'list<uint>', 'list<uint64_t>', 'list<unsigned>']:
            type_object = 'list<int>'
        if type_object in ['list<double>']:
            type_object = 'list<float>'
        for original_line in lines:
            if original_line.endswith('\n'):
                original_line = original_line[0:-1]
            line = original_line.strip()
            if not line:
                continue
            if not in_type and line.startswith('#'):
                types = line[1:].split(',')
                for type_ in types:
                    in_type = in_type or (type_object == type_.strip())
                continue
            if in_type and line.startswith('#' + serialize_types[serialize_type]):
                in_serialize = True
                continue
            if in_type and in_serialize and not in_initial_value and \
                    line.startswith('#' + serialize_init[0 if initial_value else 1]):
                in_initial_value = True
                continue
            if in_type and in_serialize and in_initial_value:
                if line.startswith('#'):
                    break
                pattern.append(original_line)
        def_ = pattern
        pattern = '\n'.join(pattern)
        # standard serialize
        pattern = pattern.replace('{', '{{')
        pattern = pattern.replace('}', '}}')
        pattern = pattern.replace('$(FIELD)', '{field}')
        pattern = pattern.replace('$(TYPE)', '{type}')
        pattern = pattern.replace('$(DEFAULT_VALUE)', '{default_value}')
        pattern = pattern.replace('$(OWNER)', '{owner}')
        pattern = pattern.replace('$(ARG_0)', '{arg_0}')
        pattern = pattern.replace('$(ARG_1)', '{arg_1}')
        pattern = pattern.replace('$(FORMAT)', '{format}')
        pattern = pattern.replace('$(NAMESPACE)', '{namespace}')
        # for map<key, value>
        pattern = pattern.replace('$(KEY_SERIALIZE)', '{key_serialize}')
        pattern = pattern.replace('$(VALUE_SERIALIZE)', '{value_serialize}')
        pattern = pattern.replace('$(KEY)', '{key}')
        pattern = pattern.replace('$(VALUE_TYPE)', '{value_type}')
        pattern = pattern.replace('$(VALUE)', '{value}')
        # for module_python:
        pattern = pattern.replace('$({{}})', '{{}}')

        if not pattern and not optional:
            print('cannot find pattern for args:')
            print(type_object, serialize_types[serialize_type], initial_value)
            print('in_type', in_type)
            print('in_serialize', in_serialize)
            print('in_initial_value', in_initial_value)
            print(def_)
            sys.exit(-1)
        return pattern

    @staticmethod
    def parse_class(cls: Class, line: str):
        from ..utils.common import smart_split
        line = line.strip()
        if cls.type in line:
            line = line[len(cls.type):]
        elif 'interface' in line:
            line = line[len('interface'):]
        line = cls.find_modifiers(line)

        pattern = re.compile(r' ([\w/]+)(<(\w+)>)*')
        parts = pattern.findall(line)[0]
        cls.name = parts[0]
        cls.superclasses = smart_split(parts[2], ',')
        cls.template_args = []
        if '/' in cls.name:
            k = cls.name.rindex('/')
            cls.group = cls.name[0:k]
            cls.name = cls.name[k + 1:]

    def parse_class_body(self, cls: Class, body: str):
        self.parse_text(body)

        for member in self.model.objects:
            if self.model.has_class(member.type):
                member.type = cls.name + member.type
            for i, arg in enumerate(member.template_args):
                if self.model.has_class(arg):
                    member.template_args[i] = cls.name + member.template_args[i]

        for func in self.model.functions:
            if self.model.has_class(func.return_type):
                func.return_type = cls.name + func.return_type
            for arg in func.args:
                if self.model.has_class(arg[1]):
                    arg[1] = cls.name + arg[1]

            for inner_class in self.model.classes:
                pattern = re.compile(r'\b{}\b'.format(inner_class.name))
                new_name = cls.name + inner_class.name
                for i, line in enumerate(func.operations):
                    func.operations[i] = re.sub(pattern, new_name, line)
        for inner_class in self.model.classes:
            inner_class_name = inner_class.name
            inner_class.name = cls.name + inner_class.name
            inner_class.group = cls.group
            inner_class.side = cls.side
            inner_class.is_test = cls.is_test
            cls.inner_classes.append(inner_class)

            for obj in self.model.objects:
                if obj.type == inner_class_name:
                    obj.type = inner_class.name
                for arg in obj.template_args:
                    if arg.type == inner_class_name:
                        arg.type = inner_class.name

            for method in self.model.functions:
                for _, arg in method.args:
                    if arg.type == inner_class_name:
                        arg.type = inner_class.name
                if method.return_type.type == inner_class_name:
                    method.return_type.type = inner_class.name

        cls.members = self.model.objects
        cls.functions = self.model.functions
        cls.user_includes = self.model.includes
        self.model.includes = set()

        for member in cls.members:
            member.set_default_initial_value()


FACTORY = '''<?php

class Factory
{
    static function build($type)
    {
        if(strcmp($type, "") == 0 || strcmp($type, "null") == 0)
        {
            return null;
        }
        try {
            require_once "$type.php";
            return new $type;
        } catch (Exception $e) {
            return null;
        }
    }
};

?>
'''

COMMON_XML = '''
function create_command_from_xml($payload)
{
    $xml     = simplexml_load_string($payload);
    $class   = $xml->getName();
    require_once "$class.php";
    $command = new $class;
    $command->deserialize_xml($xml);
    return $command;
}

function serialize_command_to_xml($command)
{
    $xml = simplexml_load_string('<'.$command->get_type().'/>');
    $command->serialize_xml($xml);
    return $xml->asXML();
}
'''

COMMON_JSON = '''
function create_command_from_json($payload)
{
    $json    = json_decode($payload);
    $class   = key($json);
    require_once "$class.php";
    $command = new $class;
    $command->deserialize_json($json->$class);
    return $command;
}

function serialize_command_to_json($command)
{
    $type = $command->get_type();
    $json = json_decode('{"'.$type.'": {}}');
    $command->serialize_json($json->$type);
    return json_encode($json, JSON_PRETTY_PRINT);
}
'''

CLONE_XML = '''
function clone_object($obj)
{
    $payload = serialize_command_to_xml($obj);
    $clone = create_command_from_xml($payload);
    return $clone;
}
'''

CLONE_JSON = '''
function clone_object($obj)
{
    $payload = serialize_command_to_json($obj);
    $clone = create_command_from_json($payload);
    return $clone;
}
'''

COMMON = '''<?php
@{xml}
@{json}

function mg_swap(&$x, &$y) {
    list($x,$y) = array($y, $x);
}

?>
'''


from enum import Enum
from .modifiers import Modifiers
from ..utils.error import Error


class AccessSpecifier(Enum):
    public = 0
    protected = 1
    private = 2

    @staticmethod
    def find_access(object_or_fucntion, string):
        if Modifiers.private in string:
            object_or_fucntion.access = AccessSpecifier.private
            string = string.replace(Modifiers.private, '')
        elif Modifiers.protected in string:
            object_or_fucntion.access = AccessSpecifier.protected
            string = string.replace(Modifiers.protected, '')
        elif Modifiers.public in string:
            object_or_fucntion.access = AccessSpecifier.public
            string = string.replace(Modifiers.public, '')
        return string

    @staticmethod
    def to_string(access):
        return {
            AccessSpecifier.public: 'public',
            AccessSpecifier.protected: 'protected',
            AccessSpecifier.private: 'private',
        }[access]


class Object(object):

    def __init__(self, type_=''):
        self.type = type_
        self.template_args = []
        self.callable_args = None
        self.name = ''
        self.initial_value = None
        self.is_pointer = False
        self.is_ref = False
        self.is_runtime = False
        self.is_static = False
        self.is_const = False
        self.is_key = False
        self.is_link = ''
        self.side = 'both'
        self.access = AccessSpecifier.public
        self.denied_intrusive = False

    def parse_type(self):
        left = self.type.find('<')
        right = -1
        try:
            right = self.type.rindex('>', left) if left != -1 else -1
        except ValueError:
            Error.exit(Error.PARSE_ERROR)
        if left > -1 and right > -1:
            args = self.type[left + 1:right].split(',')
            self.type = self.type[0:left]
            for arg in args:
                arg = arg.strip()
                self.template_args.append(arg)
        self.is_pointer = self.check_pointer()

    def check_pointer(self):
        result = '*' in self.type
        self.type = self.type.replace('*', '')
        return result or self.is_pointer

    def check_ref(self):
        result = '&' in self.type
        self.type = self.type.replace('&', '')
        return result or self.is_ref

    def find_modifiers(self, string):
        args = ''
        left = string.find('<')
        right = string.rfind('>')
        if left != -1 and right != -1:
            args = string[left:right + 1]
            string = string[0:left] + string[right + 1:]

        if Modifiers.server in string:
            self.side = Modifiers.side_server
        if Modifiers.client in string:
            self.side = Modifiers.side_client
        self.is_runtime = self.is_runtime or Modifiers.runtime in string
        self.is_static = self.is_static or Modifiers.static in string
        self.is_key = self.is_key or Modifiers.key in string
        self.is_link = self.is_link or Modifiers.link in string
        self.is_const = self.is_const or Modifiers.const in string
        self.is_const = self.is_const or self.is_link
        self.is_pointer = self.is_pointer or Modifiers.pointer in string
        self.is_ref = self.is_ref or Modifiers.ref in string
        if self.is_link:
            self.is_pointer = True

        string = AccessSpecifier.find_access(self, string)

        string = string.replace(Modifiers.server, '')
        string = string.replace(Modifiers.client, '')
        string = string.replace(Modifiers.runtime, '')
        string = string.replace(Modifiers.const, '')
        string = string.replace(Modifiers.static, '')
        string = string.replace(Modifiers.key, '')
        string = string.replace(Modifiers.link, '')
        string = string.replace(Modifiers.ref, '')
        string = string.replace(Modifiers.pointer, '')

        if args:
            string = string[0:left] + args + string[left:]

        return string

    def set_default_initial_value(self):
        if self.initial_value is None:
            if self.type in ['int', 'int64_t', 'uint', 'uint64_t', 'unsigned']:
                self.initial_value = "0"
            elif self.type in ['float', 'double']:
                self.initial_value = "0.0"
            elif self.type == 'bool':
                self.initial_value = "false"
            elif self.type == 'string':
                self.initial_value = '""'
            elif self.is_pointer:
                self.initial_value = "nullptr"


class Objects(object):
    VOID = Object('void')
    BOOL = Object('bool')
    INT = Object('int')
    STRING = Object('string')
    FLOAT = Object('float')



from ..base import WriterBase
from ..core.object import AccessSpecifier
from .serializer import Serializer


class Writer(WriterBase):

    def __init__(self, out_directory):
        WriterBase.__init__(self, out_directory)

    def write_class(self, cls):
        self.set_initial_values(cls)

        declaration_list = ''
        initialization_list = ''
        for member in cls.members:
            declare, init = self.write_object(member)
            declaration_list += declare + '\n'
            if init:
                initialization_list += init + '\n'

        functions = ''
        for method in cls.functions:
            text = self.write_function(method)
            functions += text

        imports = 'require_once "common.php";\n'
        name = cls.name
        extend = ''
        include_patter = '\nrequire_once "{}.php";'
        if cls.superclasses:
            extend = ' extends ' + cls.superclasses[0].name
            imports += include_patter.format(cls.superclasses[0].name)
        for obj in cls.members:
            if self.model.has_class(obj.type):
                if obj.type != cls.name:
                    imports += include_patter.format(obj.type)
            elif obj.type in ['list', 'map']:
                for arg in obj.template_args:
                    if self.model.has_class(arg.type) and arg.type != cls.name:
                        imports += include_patter.format(arg.type)
        imports += include_patter.format('Factory')
        if 'DataStorage' in functions:
            imports += include_patter.format('DataStorage')

        constructor_args, constructor_body = self.get_constructor_data(cls)
        out = PATTERN_FILE.format(name=name,
                                  extend=extend,
                                  declarations=declaration_list,
                                  initialize_list=initialization_list,
                                  functions=functions,
                                  imports=imports,
                                  constructor_args=constructor_args,
                                  constructor_body=constructor_body)
        return [
            ('%s.php' % cls.name, self.prepare_file(out))
        ]

    def write_object(self, obj):
        out_init = ''
        value = obj.initial_value
        cls_type = self.model.get_class(obj.type) if self.model.has_class(obj.type) else None
        if (value in [None, '"NONE"'] and not obj.is_pointer) or (cls_type and cls_type.type == 'enum'):
            if obj.type == "string":
                value = '""'
            elif obj.type == "int":
                value = "0"
            elif obj.type == "float":
                value = "0"
            elif obj.type == "uint":
                value = "0"
            elif obj.type == "bool":
                value = "false"
            elif obj.type == "list":
                value = "array()"
            elif obj.type == "map":
                value = "array()"
            else:
                if cls_type is not None and cls_type.type == 'enum':
                    value = None
                    if obj.initial_value:
                        initial_value = obj.initial_value.replace('::', '::$')
                    else:
                        initial_value = '{}::${}'.format(cls_type.name, cls_type.members[0].name)
                    out_init = '$this->{} = {};'.format(obj.name, initial_value)
                elif cls_type:
                    out_init = '$this->{} = new {}();'.format(obj.name, obj.type)

        if obj.is_static:
            out_declaration = AccessSpecifier.to_string(obj.access) + ' static ${0} = {1};'
        else:
            out_declaration = AccessSpecifier.to_string(obj.access) + ' ${0} = {1};'
        out_declaration = out_declaration.format(obj.name, Serializer().convert_initialize_value(value))
        return out_declaration, out_init

    def prepare_file(self, text):
        text = self.prepare_file_codestype_php(text)
        text = text.replace('::TYPE', '::$TYPE')
        text = text.replace('nullptr', 'null')
        text = text.replace('foreach(', 'foreach (')
        text = text.replace('for(', 'for (')
        text = text.replace('if(', 'if (')
        text = text.replace('  extends', ' extends')
        text = text.strip()
        return text

    def get_method_arg_pattern(self, obj):
        return '{type} ${name}={value}' if obj.initial_value is not None else '{type} ${name}'

    def get_method_pattern(self, method):
        return PATTERN_METHOD

    def get_required_args_to_function(self, method):
        return None

    def add_static_modifier_to_method(self, text):
        return 'static ' + text


PATTERN_FILE = '''<?php
{imports}

class {name} {extend}
{{
    //members:
    {declarations}
    
    public function __construct({constructor_args})
    {{
        {initialize_list}
        {constructor_body}
    }}
    
    //functions
    {functions}
}};

?>
'''

PATTERN_METHOD = '''{access} function {name}({args})
{{
    {body}
}}
'''



# -*- coding: utf-8 -*-

class Point(object):
    TYPE = "Point"
    __slots__ = ["row", "col"]

    def __init__(self, row=0, col=0):
        self.row = 0
        self.col = 0
        self.row = row
        self.col = col

    def __hash__(self):
        return id(self)

    def __eq__(self, rhs):
        result = True
        result = result and self.row == rhs.row
        result = result and self.col == rhs.col
        return result
        pass

    def __ne__(self, rhs):
        return not (self == rhs)
        pass

    def get_type(self):
        return Point.TYPE
        pass

    def serialize_xml(self, serializer):
        serializer.serialize(self.row, "row")
        serializer.serialize(self.col, "col")
        pass

    def deserialize_xml(self, serializer):
        self.row = serializer.deserialize("row", int, 0)
        self.col = serializer.deserialize("col", int, 0)
        pass

    def serialize_json(self, serializer):
        serializer.serialize(self.row, "row")
        serializer.serialize(self.col, "col")
        pass

    def deserialize_json(self, serializer):
        self.row = serializer.deserialize("row", int, 0)
        self.col = serializer.deserialize("col", int, 0)
        pass



# -*- coding: utf-8 -*-


class IntrusivePtr(object):

    def __init__(self, instance):
        object.__setattr__(self, 'instance', instance)

    def __setattr__(self, name, value):
        object.__setattr__(object.__getattribute__(self, 'instance'), name, value)

    def __getattribute__(self, name):
        return object.__getattribute__(self, 'instance').__getattribute__(name)

    def __eq__(self, rhs):
        if rhs is None:
            return
        self_instance = object.__getattribute__(self, 'instance')
        rhs_instance = object.__getattribute__(rhs, 'instance')
        return self_instance.__eq__(rhs_instance)

    def __hash__(self):
        return id(object.__getattribute__(self, 'instance'))


def make_intrusive(class_name, *args):
    return IntrusivePtr(class_name(*args))



import os
from ..utils import fileutils
from ..utils.error import Log


class SavePluginBase(object):

    def __init__(self, model):
        self.model = model
        self.streams = []

    def save_files(self, combine_to_one=True):
        if combine_to_one:
            self.__sort_files()
            self.__save_one()
        else:
            self.__save_all()

        for stream in self.streams:
            stream.close()
        self.__remove_old_files()

    # To override methods:
    def _create_combine_file_header(self):
        return None, None

    def _finalize_combine_file(self, content):
        return content

    def _is_need_save_file_on_combine(self, _):
        return False

    def _remove_includes(self, file_content):
        return file_content

    # Private methods
    def __save_one(self):
        self.__sort_files()

        combine_file, file_path = self._create_combine_file_header()
        for _, local_path, content in self.model.files:
            combine_file = self.__add_to_combine_file(combine_file, content)
            if self._is_need_save_file_on_combine(local_path):
                self.__save_file(local_path, combine_file)

        self.__save_file(file_path, combine_file)

    def __save_all(self):
        if isinstance(self.model.out_dict, dict):
            for _, local_path, content in self.model.files:
                self.model.out_dict[local_path] = content
            return

        for _, local_path, content in self.model.files:
            self.__save_file(local_path, content)

    def __save_file(self, local_path, content):
        self.model.created_files.append(local_path)
        full_path = fileutils.normalize_path(self.model.out_directory) + local_path
        exist = os.path.isfile(full_path)
        result, stream = fileutils.write(full_path, content)
        if result:
            self.streams.append(stream)
            msg = ' Create: {}' if not exist else ' Overwriting: {}'
            Log.message(msg.format(local_path))

    def __sort_files(self):
        def sort_func(data):
            weight = ''
            cls = data[0]
            if cls is None:
                return weight
            weight = cls.name
            while cls.superclasses:
                cls = cls.superclasses[0]
                weight = '~' + weight
            return weight
        self.model.files = sorted(self.model.files, key=sort_func)

    def __add_to_combine_file(self, combine, data):
        combine += self._remove_includes(data)
        return combine

    def __remove_old_files(self):
        files = fileutils.get_files_list(self.model.out_directory)
        for local_path in files:
            if local_path not in self.model.created_files and not local_path.endswith('.pyc'):
                os.remove(self.model.out_directory + local_path)
                Log.debug(' Removed: {}'.format(local_path))



from ..base.serializer_base import SerializerBase
from ..core.object import Object, Objects
from .protocols import JS_JSON
from .regex import RegexPatternJs


class Serializer(SerializerBase):

    def __init__(self):
        SerializerBase.__init__(self)

    def get_protocol(self, serialize_format):
        assert serialize_format == 'json'
        return JS_JSON

    def create_serialization_function(self, cls, serialize_type, serialize_format):
        return SerializerBase.create_serialization_function(self, cls, serialize_type, serialize_format)

    def get_parent_serialize_format(self):
        return '{0}.prototype.{1}.call(this, {2});\n'

    def build_map_serialization(self, obj_name, obj_template_args, serialization_type, serialize_format):
        key = obj_template_args[0]
        value = obj_template_args[1]
        assert isinstance(key, Object) and isinstance(value, Object)
        key_type = key.type
        value_type = value.type
        pattern = self.serialize_protocol[serialization_type]['map'][0]
        value_is_pointer = value.is_pointer

        def get_create_type_operation(type_):
            types = {
                'list': '[]',
                'map': '{}',
            }
            return types[type_] if type_ in types else 'new ' + type_

        if value_type not in self.model.simple_types:
            value_declaration = 'let value = {}();'.format(get_create_type_operation(value_type))
        else:
            value_declaration = ''
        key_serialize = self.build_serialize_operation_('key', key_type, None, serialization_type, key.template_args,
                                                        False, '', key.is_link, serialize_format)
        value_serialize = self.build_serialize_operation_('value', value_type, None, serialization_type,
                                                          value.template_args,
                                                          value_is_pointer, '', False, serialize_format)
        key_serialize = key_serialize.split('\n')
        for index, line in enumerate(key_serialize):
            key_serialize[index] = line
        key_serialize = '\n'.join(key_serialize)
        value_serialize = value_serialize.split('\n')
        for index, line in enumerate(value_serialize):
            value_serialize[index] = line
        value_serialize = '\n'.join(value_serialize)
        return pattern.format(field=obj_name,
                              key_serialize=key_serialize,
                              value_serialize=value_serialize,
                              key='{}',
                              owner='this.',
                              value=value_declaration,
                              value_type=value.type) + '\n'

    def convert_initialize_value(self, value):
        assert value is None or isinstance(value, str)

        if value in [None, 'nullptr', 'None']:
            value = 'null'
        value = value.replace('::', '.')
        value = RegexPatternJs.INITIALIZE[0].sub(RegexPatternJs.INITIALIZE[1], value)
        return value

    def get_serialization_function_args(self, serialize_type, serialize_format):
        return ['json', Objects.VOID]

    def build_serialize_operation(self, obj, serialization_type, serialize_format):
        return self.build_serialize_operation_(obj.name, obj.type, obj.initial_value,
                                               serialization_type, obj.template_args, obj.is_pointer, 'this.',
                                               obj.is_link, serialize_format)

    def finalize_serialize_operation(self, string):
        return string + '\n'



import setuptools
import mlc_tools.version

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mlc-tools",
    version=mlc_tools.version,
    author="Vladimir Tolmachev",
    author_email="tolm_vl@hotmail.com",
    description="A tool to generate and translate C++ code to other languages from one code-base.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mlc-tools/mlc-tools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'PyYaml'
    ],
    entry_points={
        'console_scripts': [
            'mlc = mlc_tools.console.console:main'
        ]
    },
    python_requires='>=3.6',
)




class Language(object):

    def __init__(self, module_lang, model):
        from pydoc import locate
        module_name = 'module_%s' % (module_lang if module_lang != 'py' else 'python')

        self.generator = locate('mlc_tools.%s.Generator' % module_name)()
        self.translator = locate('mlc_tools.%s.Translator' % module_name)()
        self.serializer = locate('mlc_tools.%s.Serializer' % module_name)()
        self.writer = locate('mlc_tools.%s.Writer' % module_name)(model.out_directory)
        self.save_plugin = locate('mlc_tools.%s.SavePlugin' % module_name)(model)
        self.writer.model = model
        self.writer.serializer = self.serializer

        try:
            self.registrar = locate('mlc_tools.%s.registrar' % module_name).Registrar()
        except AttributeError:
            self.registrar = None
        except TypeError:
            self.registrar = None

    def get_generator(self):
        return self.generator

    def get_translator(self):
        return self.translator

    def get_serializer(self):
        return self.serializer

    def get_writer(self):
        return self.writer



# -*- coding: utf-8 -*-
from tests.test_serializer.py.Meta import Meta
from tests.test_serializer.py.gen.BaseEnum import BaseEnum
from tests.test_serializer.py.gen.DataWrapper import DataWrapper
from tests.test_serializer.py.gen.IntrusivePtr import IntrusivePtr, make_intrusive
from tests.test_serializer.py.gen.Factory import Factory


class DeserializerXml(object):
    def __init__(self, node):
        self.node = node

    def deserialize(self, key, meta, default_value=None):
        node = self.get_child(key)
        if meta.__base__ == BaseEnum:
            value = self.deserialize_attr(key, str, '')
            return getattr(meta, value) if hasattr(meta, value) else default_value
        if isinstance(meta, Meta):
            if meta.args[0] == dict:
                return node.deserialize_dict('', meta)
            if meta.args[0] == list:
                return node.deserialize_list('', meta)
            if meta.args[0] == DataWrapper:
                from tests.test_serializer.py.gen.DataStorage import DataStorage
                value = self.deserialize_attr(key, str, '')
                return getattr(DataStorage.shared(), 'get' + meta.args[1].TYPE)(value)
            if meta.args[0] == IntrusivePtr:
                if node is not None and node.node is not None:
                    obj = Factory.build(node.node.attrib['type'])
                    obj.deserialize_xml(DeserializerXml(node.node))
                    return obj
                return None
        if hasattr(meta, 'deserialize_xml'):
            obj = meta()
            obj.deserialize_xml(node)
            return obj
        return self.deserialize_attr(key, meta, default_value)

    def deserialize_attr(self, key, meta, default_value):
        value = None
        if key and self.node is not None:
            if key in self.node.attrib:
                value = self.node.attrib[key]
            else:
                value = default_value
        if meta == bool:
            if isinstance(value, bool):
                return value
            if value is not None:
                value = value.lower()
            return True if value in ['true', 'yes'] else False if value in ['false', 'no'] else default_value if default_value else False
        if meta == str and not value and not default_value:
            return ''
        return meta(value or default_value or 0)

    def deserialize_dict(self, key, meta):
        node = DeserializerXml(self.node) if not key else self.get_child(key)
        if node.node is None:
            return {}
        result = {}
        for item in node.node:
            k = DeserializerXml(item).deserialize_key(meta.args[1])
            v = DeserializerXml(item).deserialize_value(meta.args[2])
            result[k] = v
        return result

    def deserialize_list(self, key, meta):
        if self.node is None:
            return []
        result = []
        for child in self.node:
            item = DeserializerXml(child).deserialize_list_item(meta.args[1])
            result.append(item)
        return result

    def get_child(self, key):
        return DeserializerXml(self.node.find(key) if (key and self.node is not None) else self.node)

    def deserialize_key(self, meta: Meta):
        return self.deserialize('key', meta)

    def deserialize_value(self, meta: Meta):
        return self.deserialize('value', meta)

    def deserialize_list_item(self, meta):
        if meta.__base__ == BaseEnum:
            value = self.deserialize_attr('value', str, '')
            return getattr(meta, value)
        elif hasattr(meta, 'serialize_xml'):
            obj = meta()
            obj.deserialize_xml(self)
            return obj
        elif isinstance(meta, list) or (isinstance(meta, Meta) and meta.args[0] == list):
            return self.deserialize_list('', meta)
        elif isinstance(meta, dict):
            return self.deserialize_dict('', meta)
        elif isinstance(meta, Meta):
            if meta.args[0] == DataWrapper:
                from tests.test_serializer.py.gen.DataStorage import DataStorage
                value = self.deserialize_attr('value', str, '')
                return getattr(DataStorage.shared(), 'get' + meta.args[1].TYPE)(value)
            if meta.args[0] == IntrusivePtr:
                obj = Factory.build(self.node.tag)
                obj.deserialize_xml(self)
                return obj
        else:
            return self.deserialize_attr('value', meta, None)



from ..base.generator_operator_equals import GeneratorOperatorEqualsBase


class GeneratorOperatorEquals(GeneratorOperatorEqualsBase):

    def __init__(self):
        GeneratorOperatorEqualsBase.__init__(self)

    def get_equal_method_name(self):
        return 'operator =='

    def get_not_equal_method_name(self):
        return 'operator !='

    def get_compare_method_pattern(self, cls, member):
        body_line = 'result = result && this->{0} == rhs.{0};'
        body_line_ptr = \
            'result = result && ((this->{0} == rhs.{0}) || ' \
            '(this->{0} != nullptr && rhs.{0} != nullptr && *this->{0} == *rhs.{0}));'
        return body_line_ptr if member.is_pointer or member.is_link else body_line

    def get_not_equal_method_operation(self):
        return 'return !(*this == rhs);'



import unittest

import os
import sys
import inspect

sys.path.insert(0, os.path.abspath(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/..'))
from mlc_tools.core.object import Object, Objects
from mlc_tools.core.class_ import *
from mlc_tools.core.function import *
from mlc_tools.base.model import Model
from mlc_tools.base.parser import Parser
from mlc_tools.base.language import Language
from mlc_tools.base.linker import Linker
from mlc_tools.base.validator import Validator


def create_test_model(with_ctr=False):
    cls = Class()
    cls.name = 'Test'

    m = Object()
    m.type = 'int'
    m.name = 'int_value'
    m.initial_value = '42'
    cls.members.append(m)

    m = Parser.create_object('list<string>:static test_list')
    cls.members.append(m)

    f = Function()
    f.name = 'foo'
    f.return_type = Object()
    f.return_type.type = 'int'
    f.args = [['a0', Objects.INT], ['a1', Objects.STRING]]
    f.operations = [
        'auto result = this->int_value;\n'
        'result += a0;\n'
        'return result;\n'
    ]
    cls.functions.append(f)

    if with_ctr:
        ctr = Function()
        ctr.name = 'const'
        ctr.name += 'ructor'
        ctr.return_type = Object()
        ctr.args = []
        ctr.operations = ['this->int_value = 1;']
        cls.functions.append(ctr)
        cls.generate_constructor()

    model = Model()
    # model.parser = Parser(model)
    model.add_class(cls)
    model.out_directory = '/Users/stereo7/mlc_tools_temp/'
    model.generate_tests = False
    model.generate_factory = False
    model.generate_intrusive = False
    model.generate_ref_counter = True
    model.out_dict = {}
    return model


def create_lang(lang, model):
    model.language = lang
    language = Language(lang, model)
    language.get_generator().generate(model)

    Linker().link(model)
    Validator().validate(model)

    language.get_translator().translate(model)
    return language


def save(lang, model):
    lang = create_lang(lang, model)
    lang.get_writer().save(model)
    lang.save_plugin.save_files()


def save_object(lang, model):
    cls = model.get_class('Test')
    writer = create_lang(lang, model).get_writer()
    writer.current_class = cls
    return writer.write_object(cls.members[0])


def save_function(lang, model):
    cls = model.get_class('Test')
    writer = create_lang(lang, model).get_writer()
    writer.current_class = cls
    return writer.write_function(cls.functions[0])


class TestWriteClass(unittest.TestCase):

    def test_cpp(self):
        model = create_test_model()
        model.generate_ref_counter = True
        save('cpp', model)

        self.assertTrue('Test.h' in model.out_dict)
        self.assertTrue('Test.cpp' in model.out_dict)

        hpp = model.out_dict['Test.h']
        cpp = model.out_dict['Test.cpp']
        self.assertEqual(cpp, get_cpp(False))
        self.assertEqual(hpp, get_hpp())

    def test_cpp_with_ctr(self):
        model = create_test_model(True)
        save('cpp', model)

        self.assertTrue('Test.h' in model.out_dict)
        self.assertTrue('Test.cpp' in model.out_dict)

        hpp = model.out_dict['Test.h']
        cpp = model.out_dict['Test.cpp']
        self.assertEqual(cpp, get_cpp(True))
        self.assertEqual(hpp, get_hpp())


class TestWriteObject(unittest.TestCase):

    def test_cpp(self):
        model = create_test_model()
        result = save_object('cpp', model)
        self.assertEqual(result, ('int int_value;', 'int_value(42)', ''))

    def test_python(self):
        model = create_test_model()
        result, imports = save_object('py', model)
        self.assertEqual(result, 'self.int_value = 42')
        self.assertEqual(imports, [])

    def test_php(self):
        model = create_test_model()
        result = save_object('php', model)
        self.assertEqual(result, ('public $int_value = 42;', ''))


class TestWriteFunction(unittest.TestCase):

    def test_cpp(self):
        model = create_test_model()
        result = save_function('cpp', model)
        hpp = 'int foo(int a0, const std::string& a1);\n'
        cpp = '''int Test::foo(int a0, const std::string& a1)
        {
        auto result = this->int_value;
result += a0;
return result;

        }
        \n        '''
        self.assertEqual(result[0], hpp)
        self.assertEqual(result[1], cpp)

    def test_python(self):
        model = create_test_model()
        result = save_function('py', model)
        self.assertEqual(result, '\n    def foo(self, a0, a1):\n        result = self.int_value\n        result += a0\n        return result\n        \n')

    def test_php(self):
        model = create_test_model()
        result = save_function('php', model)
        self.assertEqual(result, 'public function foo(int $a0, string $a1)\n{\n    $result = $this->int_value;\n$result += $a0;\nreturn $result;\n\n}\n')


class TestCppWriterBuildIncludesWithGroups(unittest.TestCase):

    def test_build_include(self):
        from mlc_tools.module_cpp.writer import Writer
        cls1 = Class()
        cls1.name = 'cls1'
        cls2 = Class()
        cls2.name = 'cls2'

        cls1.group = ''
        cls2.group = ''
        self.assertEqual(Writer.get_include_path_to_class(cls1, cls2), '"cls2.h"')
        cls1.group = 'a'
        cls2.group = ''
        self.assertEqual(Writer.get_include_path_to_class(cls1, cls2), '"../cls2.h"')
        cls1.group = ''
        cls2.group = 'a'
        self.assertEqual(Writer.get_include_path_to_class(cls1, cls2), '"a/cls2.h"')
        cls1.group = 'a'
        cls2.group = 'a'
        self.assertEqual(Writer.get_include_path_to_class(cls1, cls2), '"cls2.h"')
        cls1.group = 'a'
        cls2.group = 'b'
        self.assertEqual(Writer.get_include_path_to_class(cls1, cls2), '"../b/cls2.h"')

    def test_get_path_to_root(self):
        from mlc_tools.module_cpp.writer import Writer
        cls = Class()
        cls.name = 'cls'

        cls.group = 'b'
        self.assertEqual(Writer.get_path_to_root(cls), '../')
        cls.group = ''
        self.assertEqual(Writer.get_path_to_root(cls), '')

    def test_get_includes_for_header_cpp(self):
        from mlc_tools.module_cpp.writer import Writer
        # get_includes_for_header
        cls = Class()
        cls.name = 'cls'

        method = Parser.create_function('function void assertInMap(int key, map<int, int>:const:ref map)')
        cls.functions.append(method)

        writer = Writer('')
        writer.model = Model()
        includes, f, f_out = writer.get_includes_for_header(cls)
        self.assertIn('std::map', includes, '<map> in includes')


class TestCppWriterWriteObject(unittest.TestCase):

    def test_0(self):
        from mlc_tools.module_cpp.writer import Writer
        model = create_test_model()
        lang = create_lang('cpp', model)
        writer = lang.get_writer()

        member = Parser.create_object('DataBase*:static db')
        model.classes[0].members.append(member)

        self.assertEqual(writer.write_member_declaration(member), 'static intrusive_ptr<DataBase> db;')
        self.assertEqual(writer.write_member_static_init(model.classes[0], member),
                         'intrusive_ptr<DataBase> Test::db(nullptr);')


class TestWriterPrepareFile(unittest.TestCase):

    def test_xml(self):
        from mlc_tools.base.model import SerializeFormat
        writer = self.create_writer()
        writer.model.serialize_formats = SerializeFormat.xml
        text = writer.prepare_file(TestWriterPrepareFile.get_text())
        self.assertIn('xml_functional', text)
        self.assertNotIn('json_functional', text)
        self.assertNotIn('both_functional', text)
        self.assertNotIn('format=', text)

    def test_json(self):
        from mlc_tools.base.model import SerializeFormat
        writer = self.create_writer()
        writer.model.serialize_formats = SerializeFormat.json
        text = writer.prepare_file(TestWriterPrepareFile.get_text())
        self.assertNotIn('xml_functional', text)
        self.assertIn('json_functional', text)
        self.assertNotIn('both_functional', text)
        self.assertNotIn('format=', text)

    def test_both(self):
        from mlc_tools.base.model import SerializeFormat
        writer = self.create_writer()
        writer.model.serialize_formats = SerializeFormat.xml | SerializeFormat.json
        text = writer.prepare_file(TestWriterPrepareFile.get_text())
        self.assertNotIn('xml_functional', text)
        self.assertNotIn('json_functional', text)
        self.assertIn('both_functional', text)
        self.assertNotIn('format=', text)

    @staticmethod
    def create_writer():
        from mlc_tools.base.writer_base import WriterBase
        writer = WriterBase('')
        writer.model = Model()
        writer.model.generate_ref_counter = True
        return writer

    @staticmethod
    def get_text():
        text = '''
        {{format=xml}}xml_functional{{end_format=xml}}
        {{format=json}}json_functional{{end_format=json}}
        {{format=both}}both_functional{{end_format=both}}
        '''
        return text


def get_cpp(with_ctr):
    return '''#include "intrusive_ptr.h"
#include "mg_Factory.h"
#include "Test.h"
#include <string>
#include "mg_extensions.h"
#include "SerializerJson.h"
#include "SerializerXml.h"

namespace mg
{
    std::vector<std::string> Test::test_list;
    const std::string Test::TYPE("Test");

    REGISTRATION_OBJECT(Test);

    Test::Test()
    : int_value(42)
    , _reference_counter(1)
    {
{ctr}
    }

    Test::~Test()
    {
    }

    int Test::foo(int a0, const std::string& a1)
    {
        auto result = this->int_value;
        result += a0;
        return result;

    }

    bool Test::operator ==(const Test& rhs) const
    {
        bool result = true;
        result = result && this->int_value == rhs.int_value;
        return result;
    }

    bool Test::operator !=(const Test& rhs) const
    {
        return !(*this == rhs);
    }

    void Test::retain()
    {
        this->_reference_counter += 1;
    }

    int Test::release()
    {
        this->_reference_counter -= 1;
        auto counter = this->_reference_counter;
        if(counter == 0)
        delete this;
        return counter;
    }

    std::string Test::get_type() const
    {
        return Test::TYPE;
    }

} //namespace mg
'''.replace('{ctr}', '        this->int_value = 1;' if with_ctr else '')


def get_hpp():
    return '''#ifndef __mg_Test_h__
#define __mg_Test_h__

#include <cstdint>
#include "intrusive_ptr.h"
#include "pugixml/pugixml.hpp"
#include <string>
#include <vector>

namespace mg
{
    class SerializerXml;
    class DeserializerXml;
    class SerializerJson;
    class DeserializerJson;

    class Test
    {
    public:
        Test();
        ~Test();
        int foo(int a0, const std::string& a1);
        bool operator ==(const Test& rhs) const;
        bool operator !=(const Test& rhs) const;
        void retain();
        int release();
        std::string get_type() const;

        int int_value;
        static std::vector<std::string> test_list;
    private:
        int _reference_counter;
    public:
        static const std::string TYPE;

    };
} //namespace mg

#endif //#ifndef __mg_Test_h__
'''


def get_hpp_with_ctr():
    return ''''''


if __name__ == '__main__':
    unittest.main()



import xml.etree.ElementTree as ElementTree
import xml.dom.minidom
import json
import os
import re
from ..utils import fileutils
from ..utils.error import Error
from ..base.generator_data_storage_base import get_class_name_from_data_name
from ..base.generator_data_storage_base import get_data_list_name, get_data_name
from ..utils.error import Log


class DataParser(object):

    def __init__(self, classes, data_directory, filter_func):
        self.objects = {}
        self.format = None
        self.classes = classes
        self.filter_func = filter_func
        self.directories = [data_directory]

    def parse(self, additional_directories):
        directories = self.directories
        directories.extend(additional_directories)
        for directory in directories:
            directory = fileutils.normalize_path(directory)
            self._parse_directory(directory)
        self._validate()

    def flush(self, out_data_directory):
        buffer_ = ''
        if self.format is None:
            self.format = 'xml'
        filename = 'data.' + self.format
        if self.format == 'xml':
            buffer_ = self._flush_xml()
        elif self.format == 'json':
            buffer_ = self._flush_json()

        full_path = out_data_directory + filename
        exist = os.path.isfile(full_path)
        result, _ = fileutils.write(full_path, buffer_)
        if result:
            msg = ' Create: {}' if not exist else ' Overwriting: {}'
            Log.message(msg.format(filename))

    def _parse_directory(self, directory):
        files = fileutils.get_files_list(directory)
        for filename in files:
            full_path = directory + filename
            if self.filter_func is not None and not self.filter_func(full_path):
                continue
            if filename.endswith('.xml') and (self.format is None or self.format == 'xml'):
                self.format = 'xml'
                self._parse_xml(full_path)
            elif filename.endswith('.json') and (self.format is None or self.format == 'json'):
                self.format = 'json'
                self._parse_json(full_path)

    def _parse_xml(self, full_path):
        try:
            tree = ElementTree.parse(full_path)
            root = tree.getroot()

            def add(obj):
                if obj.tag not in self.objects:
                    self.objects[obj.tag] = []
                self._validate_type(obj.tag, full_path)
                self.objects[obj.tag].append(obj)

            if root.tag == 'data':
                for node in root:
                    add(node)
            else:
                add(root)
        except ElementTree.ParseError:
            Error.exit(Error.CANNOT_PARSE_XML, full_path)

    def _parse_json(self, full_path):
        root = json.loads(open(full_path, encoding='utf-8').read())

        if isinstance(root, dict):
            for key in root:
                self._parse_json_node(full_path, key, root)
        elif isinstance(root, list):
            for dict_ in root:
                for key in dict_:
                    self._parse_json_node(full_path, key, dict_)

    def _parse_json_node(self, full_path, key, dict_):
        name = key
        self._validate_type(key, full_path)
        if name not in self.objects:
            self.objects[name] = []
        self.objects[name].append(dict_)

    def _validate(self):
        pass

    def _validate_type(self, type_, filename=''):
        class_name = get_class_name_from_data_name(type_)
        valid = False
        for class_ in self.classes:
            if class_.name == class_name:
                valid = class_.is_storage
                break
        if not valid:
            Error.exit(Error.UNKNOWN_DATA_TYPE, type_, class_name, filename)

    def _flush_xml(self):
        root = ElementTree.Element('data')
        for type_ in self.objects:
            name = get_data_list_name(get_data_name(type_))
            node = ElementTree.SubElement(root, name)
            for obj in self.objects[type_]:
                pair = ElementTree.SubElement(node, 'pair')
                pair.attrib['key'] = obj.attrib['name']
                obj.tag = 'value'
                pair.append(obj)

        text = ElementTree.tostring(root).decode()
        return re.sub(r'>(\s+)<', '><', text)

    def _flush_json(self):
        dict_ = {}
        for key in self.objects:
            name = get_data_list_name(get_data_name(key))
            dict_[name] = []
            for object_ in self.objects[key]:
                dict_obj = {}
                for obj in object_:
                    dict_obj['key'] = object_[obj]['name']
                    dict_obj['value'] = object_[obj]
                dict_[name].append(dict_obj)
        return json.dumps(dict_, indent=2)



from ..base.parser import Parser
from ..core.class_ import Class
from ..core.object import Object, Objects, AccessSpecifier
from ..core.function import Function
from ..utils.common import *
from ..utils.error import Error

ASSERTS = {
    'this->assertTrue(': 1,
    'this->assertFalse(': 1,
    'this->assertEqual(': 2,
    'this->assertNotEqual(': 2,
    'this->assertNull(': 1,
    'this->assertNotNull(': 1,
    'this->assertInMap(': 2,
    'this->assertNotInMap(': 2,
    'this->assertInList(': 2,
    'this->assertNotInList(': 2,
    'this->assertInRange(': 3,
    'this->assertNotInRange(': 3,
}


class GeneratorUnitTestsInterface(object):

    def __init__(self):
        self.tests = []
        self.model = None

    def generate(self, model):
        self.model = model
        test_classes = []
        for cls in model.classes:
            test = self.generate_test_interface(cls)
            if test:
                test_classes.append(test)
        self.generate_base_classes()
        model.add_classes(test_classes)
        model.add_class(self.generate_all_tests_class())

    def generate_base_classes(self):
        base = BASE_CLASSES
        self.model.parser.parse_text(base)

    @staticmethod
    def get_member_name(cls_name):
        name_ = ''
        for i, char in enumerate(cls_name):
            if char.isupper() and i > 0:
                name_ += '_'
            name_ += char.lower()
        return name_

    def generate_test_interface(self, cls):
        if cls.name.find('ITest') == 0:
            return None
        class_name = 'ITest' + cls.name

        if not self.model.get_subclasses_of_class(class_name):
            return None

        test = Class()
        test.type = 'class'
        test.name = class_name
        test.group = 'tests'
        test.superclasses.append('TestCase')
        generated_functions = []
        for func in cls.functions:
            ignored = ['visit', 'accept']
            if func.name not in ignored and func.access == AccessSpecifier.public:
                self.add_method(test, 'test_' + func.name)
                generated_functions.append('test_' + func.name)

        if self.model.has_class(test.name[1:]):
            impl = self.model.get_class(test.name[1:])
            for func in impl.functions:
                if func.name.startswith('test_') and func.name not in generated_functions:
                    func.is_virtual = True
                    self.add_method(test, func.name)
                    self.generate_messages_if_empty(impl, func)

        method = Function()
        method.name = 'execute'
        method.return_type = Objects.VOID
        test.functions.append(method)
        for func in test.functions:
            name = func.name
            if name in ['initialize', 'execute']:
                continue
            method.operations.append('this->{}();'.format(name))

        self.tests.append(test)
        return test

    def add_method(self, class_, name):
        all_has_implement = True
        subclasses = self.model.get_subclasses_of_class(class_.name)
        for cls in subclasses:
            method = cls.get_method_with_name(name)
            if not method or method.is_abstract:
                all_has_implement = False
                Error.warning(Error.WARNING_TEST_CLASS_NOT_IMPLEMENT_METHOD, class_.name, name)
                break

        method = Function()
        method.name = name
        method.return_type = Objects.VOID
        method.is_abstract = all_has_implement
        class_.functions.append(method)

    def generate_all_tests_class(self,):
        test_all = Class()
        test_all.type = 'class'
        test_all.name = 'RunAllTests'
        test_all.group = 'tests'

        for test in self.tests:
            if self.model.has_class(test.name[1:]):
                var_name = self.get_member_name(test.name[1:])
                member = Object()
                member.type = test.name[1:]
                member.name = var_name
                test_all.members.append(member)

        method = Function()
        method.name = 'initialize'
        method.args.append(['logger', Parser.create_object('Logger*')])
        method.return_type = Objects.VOID
        test_all.functions.append(method)
        for test in test_all.members:
            var_name = self.get_member_name(test.name)
            method.operations.append('this->{}.initialize(logger);'.format(var_name))

        method = Function()
        method.name = 'execute'
        method.return_type = Objects.BOOL
        test_all.functions.append(method)
        for test in test_all.members:
            var_name = self.get_member_name(test.name)
            method.operations.append('this->{0}.execute();'.format(var_name))

        method.operations.append('bool result = true;')
        for test in test_all.members:
            var_name = self.get_member_name(test.name)
            method.operations.append('result = result && this->{0}.result;'.format(var_name))
        method.operations.append('return result;')

        return test_all

    @staticmethod
    def generate_messages_if_empty(cls: Class, method: Function):
        for assert_method, count in ASSERTS.items():
            for i, line in enumerate(method.operations):
                if line.startswith(assert_method):
                    parts = [x.strip() for x in smart_split(line[len(assert_method):line.rfind(')')], ',', ['(', ')'])]
                    if count < len(parts):
                        continue

                    ass_method = assert_method.replace('this->', '').replace('(', '')
                    message = '"{expr} is false in {cls}::{method}:\\n     Args: {args}"'
                    message = message.format(expr=ass_method,
                                             cls=cls.name,
                                             method=method.name,
                                             args=', '.join(parts).replace('"', '\\"'))
                    pos = line.rfind(');')
                    line = line[0:pos] + ', ' + message + ');'
                    method.operations[i] = line


BASE_CLASSES = '''
class tests/Logger:test:virtual
{
    function void message(string message):abstract
    function void log(string message)
    {
        this->message(message);
    }
}

class tests/TestCase:test
{
    Logger*:runtime logger
    bool:runtime result = true

    function void initialize(Logger* logger)
    {
        this->logger = logger;
    }
    function void execute():abstract
    function void add_result(bool result, string message):private
    {
        if(this->result && !result)
        {
            this->logger->message(this->get_type() + ":");
        }
        if(!result)
        {
            this->logger->message(" - Failed: " + message);
        }
        this->result = this->result && result;
    }
    function void assertTrue(bool expression, string message="")
    {
        this->add_result(expression, message);
    }
    function void assertFalse(bool expression, string message="")
    {
        this->add_result(!expression, message);
    }
    function<T> void assertEqual(T left, T right, string message="")
    {
        this->add_result(left == right, message);
    }
    function<T> void assertNotEqual(T left, T right, string message="")
    {
        this->add_result(left != right, message);
    }
    function<T> void assertNull(T value, string message="")
    {
        this->add_result(value == nullptr, message);
    }
    function<T> void assertNotNull(T value, string message="")
    {
        this->add_result(value != nullptr, message);
    }
    function<Key, Value> void assertInMap(Key key, map<Key, Value>:const:ref map, string message="")
    {
        this->add_result(in_map(key, map), message);
    }
    function<Key, Value> void assertNotInMap(Key key, map<Key, Value>:const:ref map, string message="")
    {
        this->add_result(!in_map(key, map), message);
    }
    function<T> void assertInList(T item, list<T>:const:ref list, string message="")
    {
        this->add_result(in_list(item, list), message);
    }
    function<T> void assertNotInList(T item, list<T>:const:ref list, string message="")
    {
        this->add_result(!in_list(item, list), message);
    }
    function<T> void assertInRange(T value, T min_value, T max_value, string message="")
    {
        bool result = value >= min_value && value <= max_value;
        this->add_result(result, message);
    }
    function<T> void assertNotInRange(T value, T min_value, T max_value, string message="")
    {
        bool result = value < min_value || value > max_value;
        this->add_result(result, message);
    }
}
'''



import sys
from ..base.serializer_base import SerializerBase, DESERIALIZATION, SERIALIZATION
from ..core.object import Object, Objects
from .regex import RegexPatternPython
from .protocols import PY_XML
from .protocols import PY_JSON


class Serializer(SerializerBase):

    def __init__(self):
        SerializerBase.__init__(self)

    def get_protocol(self, serialize_format):
        protocol = PY_XML if serialize_format == 'xml' else PY_JSON
        if sys.version_info[0] == 3:
            protocol = protocol.replace('.iteritems()', '.items()')
        return protocol

    def create_serialization_function(self, cls, serialize_type, serialize_format):
        method = SerializerBase.create_serialization_function(self, cls, serialize_type, serialize_format)

        if 'import' in method.body:
            imports = []
            body = []
            lines = method.body.split('\n')
            for line in lines:
                if line.startswith('        from '):
                    if line not in imports:
                        imports.append(line)
                elif line.strip():
                    body.append(line)
            method.body = '\n'.join(imports) + '\n' + '\n'.join(body)

        use_factory = RegexPatternPython.FACTORY.search(method.body) is not None
        use_data_storage = 'DataStorage.shared()' in method.body
        imports = ''
        if serialize_type == DESERIALIZATION:
            imports += '        from .Meta import Meta\n'
            for member in cls.members:
                if self.model.has_class(member.type) and member.type != cls.name:
                    line = '        from .{0} import {0}\n'.format(member.type)
                    if line not in imports and line not in method.body:
                        imports += line
            if use_factory:
                imports += '        from .Factory import Factory\n'
            if use_data_storage:
                imports += '        from .DataStorage import DataStorage\n'
        if imports and method.body:
            method.body = imports + method.body
        elif not method.body:
            method.body = '        pass'
        method.body = method.body.replace('::', '.')
        return method.body

    def get_parent_serialize_format(self):
        return '        {}.{}(self, {})\n'

    def build_map_serialization(self, obj_name, obj_template_args, serialization_type, serialize_format):
        key = obj_template_args[0]
        value = obj_template_args[1]
        assert isinstance(key, Object) and isinstance(value, Object)
        key_type = key.type
        value_type = value.type
        string = self.model.serialize_protocol[serialization_type]['map'][0]
        key_serialize = self.build_serialize_operation_('key', key_type, None, serialization_type,
                                                        key.template_args, False, '', key.is_link, serialize_format)
        value_serialize = self.build_serialize_operation_('value', value_type, None, serialization_type,
                                                          value.template_args, value.is_pointer, '', value.is_link,
                                                          serialize_format)
        key_serialize = key_serialize.split('\n')
        for index, line in enumerate(key_serialize):
            key_serialize[index] = '    ' + line
        key_serialize = '\n'.join(key_serialize)
        value_serialize = value_serialize.split('\n')
        for index, line in enumerate(value_serialize):
            value_serialize[index] = '    ' + line
        value_serialize = '\n'.join(value_serialize)
        return string.format(field=obj_name,
                             key_serialize=key_serialize,
                             value_serialize=value_serialize,
                             owner='self.') + '\n'

    def convert_initialize_value(self, value):
        assert value is None or isinstance(value, str)
        if value is None:
            return None

        if value == 'true':
            return 'True'
        if value == 'false':
            return 'False'
        if value == 'nullptr':
            return 'None'
        if '::' in value:
            value = value.replace('::', '.')
        return value

    def get_serialization_function_args(self, serialize_type, serialize_format):
        return ['serializer', Objects.VOID]

    def build_serialize_operation(self, obj, serialization_type, serialize_format):
        if serialization_type == SERIALIZATION:
            if obj.initial_value is not None:
                value = self.convert_initialize_value(obj.initial_value)
                return '\n        serializer.serialize(self.{0}, "{0}", {1})'.format(obj.name, value)
            else:
                return '\n        serializer.serialize(self.{0}, "{0}")'.format(obj.name)
        if serialization_type == DESERIALIZATION:
            meta, imports = self.create_meta_class(obj)
            imports = '\n'.join(['        from .{0} import {0}'.format(x) for x in imports if x != self.current_class.name]) if imports else ''
            if obj.initial_value:
                value = self.convert_initialize_value(obj.initial_value)
                return '\n{2}\n        self.{0} = serializer.deserialize("{0}", {1}, {3})'.format(obj.name, meta, imports, value)
            else:
                return '\n{2}\n        self.{0} = serializer.deserialize("{0}", {1})'.format(obj.name, meta, imports)
        return self.build_serialize_operation_(obj.name, obj.type, obj.initial_value,
                                               serialization_type, obj.template_args, obj.is_pointer, 'self.',
                                               obj.is_link, serialize_format)

    def finalize_serialize_operation(self, string):
        for pattern in RegexPatternPython.PEP8:
            string = pattern[0].sub(pattern[1], string)
        return '        ' + string + '\n'

    def create_meta_class(self, obj: Object) -> (str, []):
        types = {
            'string': 'str',
            'map': 'dict',
            'int': 'int',
            'uint': 'int',
            'unsigned': 'int',
            'int64_t': 'int',
            'uint64_t': 'int',
            'float': 'float',
            'double': 'float',
            'bool': 'bool',
        }
        if obj.is_link:
            return 'Meta(DataWrapper, {})'.format(obj.type), ['DataWrapper', obj.type]
        elif obj.is_pointer:
            return 'Meta(IntrusivePtr, {})'.format(obj.type), ['IntrusivePtr', obj.type]
        elif not obj.template_args:
            if obj.type in types:
                return types[obj.type], []
            return obj.type, [obj.type]
        args = [self.create_meta_class(x) for x in obj.template_args]
        type = types[obj.type] if obj.type in types else obj.type
        meta = 'Meta({}, {})'.format(type, ', '.join([x[0] for x in args]))
        args_ = []
        for x in args:
            if x[1]:
                args_.extend(x[1])
        args = [obj.type] if self.model.has_class(obj.type) else []
        args.extend(args_)
        return meta, args



# -*- coding: utf-8 -*-
from tests.test_serializer.py.gen.IntrusivePtr import make_intrusive


class Factory(object):

    @staticmethod
    def build(type):
        if type == "AllTypesChildren":
            from .AllTypesChildren import AllTypesChildren
            return make_intrusive(AllTypesChildren)
        if type == "AllTypes":
            from .AllTypes import AllTypes
            return make_intrusive(AllTypes)
        if type == "DataUnit":
            from .DataUnit import DataUnit
            return make_intrusive(DataUnit)
        if type == "VisualUnit":
            from .VisualUnit import VisualUnit
            return make_intrusive(VisualUnit)
        return None



from ..core.function import Function
from ..core.object import Objects


class GeneratorOperatorValueOf(object):

    def __init__(self):
        self.model = None

    def generate(self, model):
        self.model = model
        for cls in model.classes:
            GeneratorOperatorValueOf.add_value_of(cls)

    @staticmethod
    def add_value_of(class_):
        if class_.is_inline:
            return

        method = Function()
        method.name = 'valueOf'
        method.return_type = Objects.STRING
        method.operations.append('return serialize_command_to_json(this);')
        class_.functions.append(method)



from ..core.function import Function
from ..core.object import Object, Objects


class GeneratorOperatorEqualsBase(object):

    def __init__(self):
        self.model = None

    def generate(self, model):
        self.model = model
        for cls in model.classes:
            if cls.type == 'enum':
                continue
            if cls.is_inline:
                continue
            if cls.name in ['DataStorage']:
                continue
            self.add_equal_method(cls)
            self.add_not_equal_method(cls)

    @staticmethod
    def get_const_ref(cls_):
        ref = Object()
        ref.type = cls_.name
        ref.is_const = True
        ref.is_ref = True
        return ref

    def add_equal_method(self, cls):
        operator = Function()
        operator.name = self.get_equal_method_name()
        operator.return_type = Objects.BOOL
        operator.args.append(['rhs', GeneratorOperatorEqualsBase.get_const_ref(cls)])
        operator.is_const = True

        superclass = self.model.get_class(cls.superclasses[0]) if cls.superclasses else None
        if not superclass or superclass.is_inline:
            operator.operations.append('bool result = true;')
        else:
            pattern = self.get_call_superclass_equal()
            operator.operations.append('bool result = ' + pattern.
                                       format(cls.superclasses[0], self.get_equal_method_name()))
        for member in cls.members:
            if member.is_static or member.is_const or member.type == 'Observable':
                continue
            skip = False
            for t in member.template_args:
                if t.type == 'Observable':
                    skip = True
                    break
            if skip:
                continue;

            pattern = self.get_compare_method_pattern(cls, member)
            operator.operations.append(pattern.format(member.name))
        operator.operations.append('return result;')
        cls.functions.append(operator)

    def add_not_equal_method(self, cls):
        operator = Function()
        operator.name = self.get_not_equal_method_name()
        operator.return_type = Objects.BOOL
        operator.args.append(['rhs', GeneratorOperatorEqualsBase.get_const_ref(cls)])
        operator.is_const = True
        operator.operations.append(self.get_not_equal_method_operation())
        cls.functions.append(operator)

    def get_equal_method_name(self):
        return ''

    def get_not_equal_method_name(self):
        return ''

    # pylint: disable=unused-argument
    def get_compare_method_pattern(self, cls, member):
        return 'result = result && this->{0} == rhs.{0};'

    def get_not_equal_method_operation(self):
        return ''

    def get_call_superclass_equal(self):
        return 'this->{}::{}(rhs);'



import subprocess


class SubprocessWrapper(object):
    VERBOSE = False

    def __init__(self, arguments, working_directory=None, require_out=False, require_log=False):
        assert isinstance(arguments, list) or isinstance(arguments, str)
        assert require_out != require_log or require_out is False

        if isinstance(arguments, str):
            arguments = arguments.split(' ')

        assert len(arguments) > 0

        self.arguments = arguments
        self.require_out = require_out
        log = require_log or SubprocessWrapper.VERBOSE
        stdout = None if log and not require_out else subprocess.PIPE
        stderr = None if log and not require_out else subprocess.PIPE
        self.process = subprocess.Popen(arguments, stdout=stdout, stderr=stderr, cwd=working_directory)
        self.code = 0

    def call(self):
        out, err = self.process.communicate()
        self.code = self.process.returncode
        if self.require_out:
            return self.code, out.decode(), err.decode()
        return self.code



from ..base.model import SerializeFormat


class GeneratorConfig(object):

    def __init__(self):
        pass

    @staticmethod
    def generate(model):
        content = ''
        filename = 'config.php'

        for serialize_format, format_string in SerializeFormat.get_all():
            support = 'true' if model.serialize_formats & serialize_format != 0 else 'false'
            line = '\n\tpublic static $SUPPORT_{}_PROTOCOL = {};'.format(format_string.upper(), support)
            content += line
        content = CONTENT.format(content)
        model.add_file(None, filename, content)


CONTENT = '''<?php

class Config {{{0}
}};

?>'''
